from copy import deepcopy
from manim import WHITE, Scene, tempconfig, Mobject
from manim_lamination_builder.custom_json import custom_dump, custom_parse
from manim_lamination_builder.lamination import Lamination
from manim_lamination_builder.points import FloatWrapper, UnitPoint
from typing import Tuple, Union
from manim_lamination_builder.animation import AnimateLamination
from manim_lamination_builder.generate import curried_colorize_with_respect_to


def remove_occluded(
    lam: Lamination, occlusion: Tuple[UnitPoint, UnitPoint]
) -> Lamination:
    ret = deepcopy(lam)
    a, b = occlusion[0].to_float(), occlusion[1].to_float()

    def criteria(point):  # weather it is not occluded
        if b > a:
            return point.to_float() < a or point.to_float() >= b
        else:
            return not (point.to_float() < b or point.to_float() >= a)

    ret.occlusion = occlusion

    ret.points = list(filter(criteria, ret.points))
    ret.polygons = list(
        filter(lambda polly: all(criteria(point) for point in polly), ret.polygons)
    )
    return ret


def morph_function(x: float, occlusion: Tuple[UnitPoint, UnitPoint]) -> float:
    a, b = occlusion[0].to_float(), occlusion[1].to_float()
    if a > b:
        a, b = b, a
    bite_length = b - a
    remaining_length = 1 - bite_length
    # Calculate the midpoint of the range
    midpoint = (a + b) / 2
    # Calculate the opposite of the midpoint
    opposite = midpoint + 0.5 if midpoint < 0.5 else midpoint - 0.5

    # Determine which side of the midpoint the angle is on
    if x >= a and x <= b:
        # The angle is in the range, so map it to the midpoint
        return midpoint
    elif x < a:
        # The angle is below the range, so stretch the lower half of the circle
        return ((x - a) / remaining_length) + midpoint
    else:
        # The angle is above the range, so stretch the upper half of the circle
        return ((x - b) / remaining_length) + midpoint


def result(lam: Lamination) -> Lamination:
    assert lam.occlusion is not None
    remaining_degree = lam.radix
    if (
        lam.occlusion[0].has_degree()
        and lam.occlusion[1].has_degree()
        and lam.occlusion[0].after_sigma().cleared()
        == lam.occlusion[1].after_sigma().cleared()
    ):
        lost_criticalitys = int(
            lam.occlusion[1].after_sigma().to_float()
            - lam.occlusion[0].after_sigma().to_float()
        )
        remaining_degree = lam.radix - lost_criticalitys

    def mapping(p: UnitPoint) -> UnitPoint:
        assert lam.occlusion is not None
        return FloatWrapper(
            morph_function(p.to_float(), lam.occlusion), remaining_degree
        )

    ret = remove_occluded(lam, occlusion=lam.occlusion).apply_function(mapping)

    ret.occlusion = None
    ret.radix = remaining_degree
    ret.colorizer = curried_colorize_with_respect_to(ret.points)

    return ret


class MorphOcclusion(AnimateLamination):
    def __init__(
        self,
        initial: Lamination,
        occlusion: Tuple[UnitPoint, UnitPoint],
        start_mobject: Union[Mobject, None] = None,
        **kwargs,
    ) -> None:
        initial.occlusion = occlusion
        reported_initial = remove_occluded(initial, occlusion)
        reported_final = result(initial)
        super().__init__(reported_initial, reported_final, start_mobject, **kwargs)


class _MyScene(Scene):
    def construct(self):
        self.camera.background_color = WHITE

        initial = custom_parse(
            """{
        "polygons": [["0_003", "0_030", "0_300"],
          ["1_003", "3_030", "3_300"],
          ["2_003", "2_030", "2_300"],
          ["3_003", "1_030", "1_300"]],
        "chords": [],
        "points": [],
        "radix": 4}"""
        )
        assert isinstance(initial, Lamination)
        initial.auto_populate()
        occlusion = (initial.polygons[0][0], initial.polygons[0][2])
        self.add(initial.build())
        self.wait(2)
        self.clear()
        self.play(MorphOcclusion(initial, occlusion, run_time=5))
        self.wait(2)


if __name__ == "__main__":
    with tempconfig({"quality": "medium_quality", "preview": True}):
        scene = _MyScene()
        scene.render()



# write a python function that takes a single argument which is a list length n of lists of length d. it should generate all of the ways of combining the elements of the lists into lists with no overlap and taking exactly one element from each of the original lists.  attempt to use combinatoric function from itertools.


from copy import deepcopy
from typing import List, Callable
from manim.utils.color import Colors
from itertools import product

from manim_lamination_builder import (
    UnitPoint,
    NaryFraction,
    Lamination,
    Chord,
    custom_dump,
    custom_parse,
)


def crosses(A: NaryFraction, B: NaryFraction, decided_on: Lamination):
    cords: List[Chord] = []

    target = Chord(A, B)
    return any([target.crosses(reference) for reference in cords])


def combine(list_of_lists):
    return list(product(*list_of_lists))


input_lists = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]


result = combine(input_lists)
print(result)
# [(1, 3, 5), (1, 3, 6), (1, 4, 5), (1, 4, 6), (2, 3, 5), (2, 3, 6), (2, 4, 5), (2, 4, 6)]



# Licensed under the The AGPLv3 License (AGPLv3)
# Copyright (c) 2023 Forrest M. Hilton <forrestmhilton@gmail.com>
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from typing import List, Callable, Tuple, Union
from manim import (
    BLUE,
    ORIGIN,
    RED,
    BLACK,
    TAU,
    Arc,
    Dot,
    Mobject,
    VMobject,
    Circle,
)
from manim.utils.color import Colors, Color
from manim_lamination_builder.points import UnitPoint
from manim_lamination_builder.chord import make_and_append_bezier


background = BLACK


class Lamination:
    polygons: List[List[UnitPoint]]
    points: List[UnitPoint]
    # occludes the region bounded by the chord and the arc from the first to the second CCW
    occlusion: Union[Tuple[UnitPoint, UnitPoint], None]
    radix: int
    colorizer: Callable[[UnitPoint], Colors]

    def __init__(
        self,
        polygons: List[List[UnitPoint]],
        points: List[UnitPoint],
        radix: int,
        colorizer=lambda p: Colors.red,
        occlusion: Union[Tuple[UnitPoint, UnitPoint], None] = None,
    ) -> None:
        self.polygons = polygons
        self.points = points
        self.radix = radix
        self.colorizer = colorizer
        self.occlusion = occlusion

    def auto_populate(self):
        for polygon in self.polygons:
            for point in polygon:
                if point not in self.points:
                    self.points.append(point)

    def build(self, radius=1.0, center=ORIGIN) -> Mobject:
        ret = Mobject()
        if self.occlusion is not None:
            delta = self.occlusion[0].to_angle() - self.occlusion[1].to_angle()
            if delta < 0:
                delta += TAU
            unit_circle = Arc(
                start_angle=self.occlusion[1].to_angle(),
                angle=delta,
                color=BLACK,
                radius=radius,
            )
        else:
            unit_circle = Circle(color=BLACK, radius=radius)  # create a circle
        unit_circle.move_arc_center_to(center)
        ret.add(unit_circle)  # show the circle on screen

        for polygon in self.polygons:
            shape = VMobject(BLUE, 1, color=BLACK)
            for i in range(len(polygon)):
                a = polygon[i]
                b = polygon[(i + 1) % len(polygon)]
                make_and_append_bezier(shape, a, b)
            ret.add(shape)

        for point in self.points:
            ret.add(
                Dot(
                    point.to_cartesian(), color=self.colorizer(point).value, radius=0.04
                )
            )

        # build a chord for occlusion

        if self.occlusion is not None:
            occlusion = VMobject(color=RED)
            make_and_append_bezier(occlusion, self.occlusion[0], self.occlusion[1])
            ret.add(occlusion)

        for submobject in ret.submobjects:
            if submobject is unit_circle:
                continue
            submobject.points *= radius
            submobject.points += center

        return ret

    def apply_function(self, f: Callable[[UnitPoint], UnitPoint]) -> "Lamination":
        new_polygons = []
        for poly in self.polygons:
            new_poly = [f(p) for p in poly]
            new_polygons.append(new_poly)
        new_points = [f(p) for p in self.points]
        return Lamination(new_polygons, new_points, self.radix,self.colorizer)



if __name__ == "__main__":
    from manim_lamination_builder import custom_dump, read_file_to_laminations, Main
    from manim import (
        tempconfig,
    )
    import sys
    import os

    file = sys.argv[-1]

    path = os.path.join(os.getcwd(), file)
    laminations = read_file_to_laminations(path)
    for lamination in laminations:
        lamination.auto_populate()

    with tempconfig({"quality": "medium_quality", "preview": True}):
        scene = Main(laminations)
        scene.render()



# Licensed under the The AGPLv3 License (AGPLv3)
# Copyright (c) 2023 Forrest M. Hilton <forrestmhilton@gmail.com>
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import List, Union
from manim_lamination_builder.lamination import Lamination
from manim_lamination_builder.points import FloatWrapper, NaryFraction
import json
import json5


class CustomEncoder(json.JSONEncoder):
    def default(self, v):
        types = {"NaryFraction": lambda v: v.to_string()}
        vtype = type(v).__name__
        if vtype == "Lamination":
            ret = v.__dict__.copy()
            ret.pop("colorizer")
            return ret
        if vtype == "Chord":
            return list(v.__dict__.values())
        if vtype == "FloatWrapper":
            return v.value
        if vtype in types:
            return types[type(v).__name__](v)
        else:
            return json.JSONEncoder.default(self, v)


class CustomDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        super().__init__(object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, dct):
        if "radix" in dct:
            radix = dct["radix"]

            def point_handler(v):
                if isinstance(v, float):
                    return FloatWrapper(v, radix)
                else:
                    return NaryFraction.from_string(radix, v)

            def list_handler(list):
                assert isinstance(list, List)
                return [*map(point_handler, list)]

            polygons = [*map(list_handler, dct.get("polygons", []))]
            points = list_handler(dct.get("points", []))
            occlusion = list_handler(dct.get("occlusion", []))
            if len(occlusion) == 0:
                occlusion = None
            else:
                assert len(occlusion) == 2
                occlusion = (occlusion[0], occlusion[1])
            return Lamination(polygons, points, radix, occlusion=occlusion)
        return dct


def preprocess_for_json5(json5_str: str) -> str:
    obj = json5.loads(json5_str)
    json_str = json.dumps(obj)
    return json_str


def read_file_to_laminations(path) -> List[Lamination]:
    with open(path) as f:
        json_str = preprocess_for_json5(f.read())
        return json.loads(json_str, cls=CustomDecoder)


def custom_dump(data_of_a_type_defined_in_this_project) -> str:
    return json.dumps(data_of_a_type_defined_in_this_project, cls=CustomEncoder)


def custom_parse(string: str) -> Union[Lamination, List[Lamination]]:
    json_str = preprocess_for_json5(string)
    return json.loads(json_str, cls=CustomDecoder)


def parse_lamination(string: str) -> Lamination:
    lam = custom_parse(string)
    assert isinstance(lam, Lamination)
    return lam



from itertools import accumulate
from typing import List, Set
from manim_lamination_builder import (
    parse_lamination,
    custom_dump,
    Lamination,
    UnitPoint,
)


def polygons_to_leafs(lam: Lamination) -> Lamination:
    leafs: List[List[UnitPoint]] = []
    for polygon in lam.polygons:
        for i in range(len(polygon)):
            leafs.append([polygon[i], polygon[(i + 1) % len(polygon)]])
    return Lamination(leafs, lam.points, lam.radix)


lam = parse_lamination(
    """
{
    "polygons": [["0_003", "0_030", "0_300"], 
      ["1_003", "3_030", "3_300"], 
      ["2_003", "2_030", "2_300"], 
      ["3_003", "1_030", "1_300"]], 
    "points": [], 
    "radix": 4
  }
"""
)


def leafs_to_polygons(lam: Lamination) -> Lamination:
    "assumes all polygons are leafs and identifies finite gaps"
    polygons: List[Set[UnitPoint]] = []
    for leaf in lam.polygons:
        used_leaf = False
        for p, other in [(leaf[0], leaf[1]), (leaf[1], leaf[0])]:
            for polygon in polygons:
                if p in polygon:
                    polygon_with_other = next(
                        filter(
                            lambda others_polygon: other in others_polygon
                            and polygon is not others_polygon,
                            polygons,
                        ),
                        None,
                    )
                    if polygon_with_other is None:
                        polygon.add(other)
                    else:
                        polygon.update(polygons.pop(polygons.index(polygon_with_other)))
                    used_leaf = True
                    break
            if used_leaf:
                break
        if not used_leaf:
            polygons.append(set(leaf))

    polygons_ = list(map(lambda s: sorted(s, key=lambda p: p.to_float()), polygons))
    return Lamination(polygons_, lam.points, lam.radix)


leafs = polygons_to_leafs(lam)
print(custom_dump(leafs))
polygons = leafs_to_polygons(leafs)
print(custom_dump(polygons))



# Licensed under the The AGPLv3 License (AGPLv3)
# Copyright (c) 2023 Forrest M. Hilton <forrestmhilton@gmail.com>
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import List

from math import cos, pi, sin, floor

from manim.animation.animation import deepcopy
from abc import ABC, abstractmethod
import numpy as np


def angle_to_cartesian(angle: float):
    return np.array([cos(angle), sin(angle), 0])


class UnitPoint(ABC):
    @abstractmethod
    def to_float(self) -> float:
        pass

    @abstractmethod
    def cleared(self) -> "UnitPoint":
        pass

    @abstractmethod
    def after_sigma(self) -> "FloatWrapper":
        pass

    def to_angle(self) -> float:
        return self.to_float() * 2 * pi

    def to_cartesian(self):
        return angle_to_cartesian(self.to_angle())

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()

    def __hash__(self):
        return hash(floor(self.to_float() / 0.0000002))

    def has_degree(self):
        return True


class FloatWrapper(UnitPoint):
    def __init__(self, value: float, degree: int = None):
        self.value = value
        self.degree = degree

    def to_float(self):
        return self.value

    def cleared(self) -> "FloatWrapper":
        return FloatWrapper(self.value % 1, self.degree)

    def after_sigma(self) -> "FloatWrapper":
        after = deepcopy(self.cleared())
        after.value *= self.degree
        return after

    def has_degree(self):
        return self.degree is not None


class NaryFraction(UnitPoint):
    def __init__(self, base: int, exact: List[int], repeating: List[int], overflow=0):
        assert base != 1
        self.base = base
        self.exact = exact
        self.repeating = repeating
        self.overflow = overflow

    def cleared(self) -> "NaryFraction":
        return NaryFraction(self.base, self.exact, self.repeating)

    @staticmethod
    def from_string(base, string_representation):
        overflow = 0
        if "." in string_representation:
            overflow, string_representation = string_representation.split(".")
            overflow = int(overflow)
        parts = string_representation.split("_")
        exact = [int(i) for i in parts[0]]
        repeating = []
        if len(parts) > 1:
            repeating = [int(i) for i in parts[1]]
        return NaryFraction(base, exact, repeating, overflow)

    def to_string(self):
        overflow_string = ""
        if self.overflow != 0:
            overflow_string = str(self.overflow) + "."
        # convert exact part to string
        exact_string = ""
        for i in self.exact:
            exact_string += str(i)

        # convert repeating part to string
        repeating_string = ""
        if self.repeating:
            repeating_string = "_"
            for i in self.repeating:
                repeating_string += str(i)

        return overflow_string + exact_string + repeating_string

    def after_sigma(self) -> "NaryFraction":
        after = deepcopy(self)
        # multiply the exact part by base
        if after.repeating:
            carry = after.repeating.pop(0)
            after.repeating.append(carry)
            after.exact.append(carry)
        after.overflow = after.exact.pop(0)
        return after

    def without_enharmonics(self):
        # TODO: refer to original to do more of this
        ret = deepcopy(self)
        while (
            len(ret.exact) != 0
            and len(ret.repeating) != 0
            and ret.repeating[-1] == ret.exact[-1]
        ):
            ret.exact.pop(-1)
            ret.repeating.insert(0, ret.repeating.pop(-1))
        # to many zeros
        while len(ret.exact) != 0 and ret.exact[-1] == 0 and sum(ret.repeating) == 0:
            ret.exact.pop(-1)

        assert (
            abs(self.to_float() - ret.to_float()) < 1e-9
        ), "Assertion Error: self not equal to ret after converting to float."
        return ret

    def pre_images(self) -> List["NaryFraction"]:
        ret = self.without_enharmonics()
        return [
            NaryFraction(self.base, [digit] + ret.exact, self.repeating)
            for digit in range(self.base)
        ]

    def siblings(self) -> List["NaryFraction"]:
        assert max(self.exact + self.repeating) < self.base
        ret = self.after_sigma().cleared().pre_images()
        assert self.cleared() in ret, "{} is not in {}".format(
            self.cleared().to_string(), list(map(lambda p: p.to_string(), ret))
        )
        return ret

    def to_float(self) -> float:
        value = sum([n / self.base ** (i + 1) for i, n in enumerate(self.exact)])
        if len(self.repeating) != 0:
            value += (
                sum(
                    [
                        n / self.base ** (i + 1 - len(self.repeating))
                        for i, n in enumerate(self.repeating)
                    ]
                )
                / self.base ** len(self.exact)
                / (self.base ** len(self.repeating) - 1)
            )
        return value + self.overflow

    def cartesian_lerp(self, other: "NaryFraction", alpha: float):
        angle = (1 - alpha) * self.to_angle() + alpha * other.to_angle()
        return angle_to_cartesian(angle)

    def after_sigma_shortest_ccw(self):
        assert self == self.cleared()
        ret = self.after_sigma().cleared()
        if ret.to_angle() < self.to_angle():
            ret = NaryFraction(ret.base, ret.exact, ret.repeating, 1)
        assert ret.to_angle() > self.to_angle()
        assert ret.to_float() - self.to_float() <= 1
        assert ret.overflow <= 1
        return ret


def sigma(p: UnitPoint):
    return p.after_sigma()


assert NaryFraction(3, [1], [1, 0, 1]).to_string() == "1_101"
assert NaryFraction(2, [1], []).to_string() == "1"

assert NaryFraction.from_string(3, "1_101") == NaryFraction(3, [1], [1, 0, 1])
assert NaryFraction.from_string(2, "1") == NaryFraction(2, [1], [])

assert NaryFraction.from_string(3, "_101").after_sigma().to_string() == "1._011"
assert NaryFraction.from_string(3, "1._101").to_string() == "1._101"
assert (
    NaryFraction.from_string(10, "_33").to_float()
    == NaryFraction.from_string(3, "1").to_float()
)


assert NaryFraction.from_string(10, "_9").to_float() == 1.0

assert NaryFraction.from_string(4, "0_300").without_enharmonics().to_string() == "_030"


# ["_302", "_322", "_023", "_223"]
# ["_230", "_232", "_302", "_322"]

# assert NaryFraction.from_string(4,"_230").to_string() == "1._023"
# assert NaryFraction.from_string(4, "_232").to_string() == "1._223"
a = NaryFraction.from_string(4, "_230")
assert a.after_sigma().to_string() == "2._302"
assert a.to_string() == "_230"

assert a.after_sigma_shortest_ccw().to_string() == "_302"
assert a.to_string() == "_230"



from copy import deepcopy
from manim_lamination_builder.lamination import Lamination
from manim_lamination_builder.chord import Chord

from manim_lamination_builder.custom_json import custom_dump, custom_parse
from typing import List, Callable
from manim.utils.color import Colors

from manim_lamination_builder.points import UnitPoint, NaryFraction, sigma


def crosses(A: NaryFraction, B: NaryFraction, loops):
    cords: List[Chord] = []
    for loop in loops:
        for i in range(len(loop)):
            x = loop[i]
            y = loop[(i + 1) % len(loop)]
            cords.append(Chord(x, y))
    cords.pop()

    target = Chord(A, B)
    return any([target.crosses(reference) for reference in cords])


# mutates the given lists
def _generate(
    loops: List[List[NaryFraction]],
    remaining: List[List[NaryFraction]],
    original_shape: List[NaryFraction],
) -> List[List[List[NaryFraction]]]:
    order = len(original_shape)
    # must include original polygon
    if len(loops) == 0:
        loops.append([])
        for i in range(order):
            p = original_shape[i]
            assert p == remaining[i].pop(remaining[i].index(p))
            loops[0].append(p)

    assert len(loops) != 0
    # select the loop in progress
    if len(loops[-1]) == order:  # is it finished?
        if len(remaining[0]) == 0:
            return [loops]
        loops.append([remaining[0].pop(0)])  # start next loop

    sibling_portraits = []
    assert len(loops[-1]) != 0
    next_connection_options = remaining[
        len(loops[-1])
    ]  # iterate over options for its next step / how to continue polygon
    for i, next_point in enumerate(next_connection_options):
        if not crosses(next_point, loops[-1][-1], loops):  # TODO: check for crossing
            n_remaining = deepcopy(remaining)
            n_loops = deepcopy(loops)
            n_loops[-1].append(n_remaining[len(loops[-1])].pop(i))
            sibling_portraits.extend(_generate(n_loops, n_remaining, original_shape))

    return sibling_portraits


def curried_colorize_with_respect_to(
    original_shape: List[UnitPoint],
) -> Callable[[UnitPoint], Colors]:
    image = list(map(lambda p: p.after_sigma().cleared(), original_shape))

    def colorize(p: UnitPoint) -> Colors:
        colors = [
            Colors.pure_red,
            Colors.pure_green,
            Colors.pure_blue,
            Colors.yellow,
            Colors.purple,
        ]
        x = p.after_sigma().cleared()
        if x in image:
            if image.index(x) > 5:
                return Colors.black
            return colors[image.index(x)]
        return Colors.black

    return colorize


def generate_sibling_portraits(original_shape: List[NaryFraction]) -> List[Lamination]:
    "For reasons descirbed in my may 18th talk at the Nippising Topology workshop, this can be predicted using the Fuss-Catillan numbers."
    order = len(original_shape)
    degree = original_shape[0].base
    original_shape = list(map(lambda x: x.cleared(), original_shape))

    # generate all the point
    # Each index in all points contains the list of points which are sibling to the
    # corresponding point in original shape.
    # For example the points in the 1st list of all points includes the 1st
    # point in the original shape.
    all_points = []
    colorize = curried_colorize_with_respect_to(original_shape)

    for i in range(order):
        siblings = original_shape[i].siblings()
        while siblings[0].to_float() < original_shape[i].to_float():
            siblings.append(siblings.pop(0))
        all_points.append(siblings)

    portraits = _generate([], deepcopy(all_points), original_shape)
    data = [Lamination(pollygons, [], degree, colorize) for pollygons in portraits]
    for lam in data:
        lam.auto_populate()

    return data


def generate_unicritical_lamination(degree, order):
    starting_point = NaryFraction.from_string(degree, "_" + "0" * (order - 1) + "1")
    original_shape = [starting_point]
    # generate original shape
    for i in range(order - 1):
        point = original_shape[i]
        original_shape.append(point.after_sigma().cleared())

    data = generate_sibling_portraits(original_shape)
    return data


def remove_non_original_pollygons(lams: List[Lamination]):
    lam = lams[0]
    lam.polygons = [lam.polygons[0]]
    return [lam]


# generate_unicritical_lamination(4, 3)

# "0_122" is a proposed test case
